# This code is part of Kessler, a machine learning library for spacecraft collision avoidance.
#
# Copyright (c) 2020-
# Trillium Technologies
# University of Oxford
# Giacomo Acciarini (giacomo.acciarini@gmail.com)
# and other contributors, see README in root of repository.
#
# GNU General Public License version 3. See LICENSE in root of repository.


import unittest
import tempfile
import uuid
import os
import pandas as pd

from kessler import EventDataset


class EventDatasetTestCase(unittest.TestCase):
    def test_from_pandas(self):
        file_content = """
,event_id,batch_insertion_epoch,header_comment,ccsds_cdm_vers,creation_date,originator,message_id,message_for,relative_metadata_data_comment,tca,miss_distance,relative_speed,relative_position_r,relative_position_t,relative_position_n,relative_velocity_r,relative_velocity_t,relative_velocity_n,start_screen_period,stop_screen_period,screen_volume_frame,screen_volume_shape,screen_volume_x,screen_volume_y,screen_volume_z,screen_entry_time,screen_exit_time,probability,max_probability,jspoc_probability,depth_of_intrusion,t_object_metadata_comment,t_object_designator,t_catalog_name,t_object_name,t_international_designator,t_ephemeris_name,t_covariance_method,t_maneuverable,t_ref_frame,t_object_type,t_operator_id,t_orbit_center,t_gravity_model,t_atmospheric_model,t_n_body_perturbations,t_solar_rad_pressure,t_earth_tides,t_intrack_thrust,t_object_data_comment,t_object_od_params_comment,t_time_lastob_start,t_time_lastob_end,t_recommended_od_span,t_actual_od_span,t_obs_available,t_obs_used,t_tracks_available,t_tracks_used,t_residuals_accepted,t_weighted_rms,t_object_add_params_comment,t_area_pc,t_area_drg,t_area_srg,t_mass,t_cd_area_over_mass,t_cr_area_over_mass,t_thrust_acceleration,t_sedr,t_object_state_comment,t_x,t_y,t_z,t_x_dot,t_y_dot,t_z_dot,t_j2k_sma,t_j2k_ecc,t_j2k_inc,t_j2k_a_per,t_j2k_m_ano,t_j2k_raan,t_object_covar_comment,t_cr_r,t_ct_r,t_ct_t,t_cn_r,t_cn_t,t_cn_n,t_crdot_r,t_crdot_t,t_crdot_n,t_crdot_rdot,t_ctdot_r,t_ctdot_t,t_ctdot_n,t_ctdot_rdot,t_ctdot_tdot,t_cndot_r,t_cndot_t,t_cndot_n,t_cndot_rdot,t_cndot_tdot,t_cndot_ndot,t_cdrg_r,t_cdrg_t,t_cdrg_n,t_cdrg_rdot,t_cdrg_tdot,t_cdrg_ndot,t_cdrg_drg,t_csrp_r,t_csrp_t,t_csrp_n,t_csrp_rdot,t_csrp_tdot,t_csrp_ndot,t_csrp_drg,t_csrp_srp,t_cthr_r,t_cthr_t,t_cthr_n,t_cthr_rdot,t_cthr_tdot,t_cthr_ndot,t_cthr_drg,t_cthr_srp,t_cthr_thr,c_object_metadata_comment,c_object_designator,c_catalog_name,c_object_name,c_international_designator,c_ephemeris_name,c_covariance_method,c_maneuverable,c_ref_frame,c_object_type,c_operator_id,c_orbit_center,c_gravity_model,c_atmospheric_model,c_n_body_perturbations,c_solar_rad_pressure,c_earth_tides,c_intrack_thrust,c_object_data_comment,c_object_od_params_comment,c_time_lastob_start,c_time_lastob_end,c_recommended_od_span,c_actual_od_span,c_obs_available,c_obs_used,c_tracks_available,c_tracks_used,c_residuals_accepted,c_weighted_rms,c_object_add_params_comment,c_area_pc,c_area_drg,c_area_srg,c_mass,c_cd_area_over_mass,c_cr_area_over_mass,c_thrust_acceleration,c_sedr,c_object_state_comment,c_x,c_y,c_z,c_x_dot,c_y_dot,c_z_dot,c_j2k_sma,c_j2k_ecc,c_j2k_inc,c_j2k_a_per,c_j2k_m_ano,c_j2k_raan,c_object_covar_comment,c_cr_r,c_ct_r,c_ct_t,c_cn_r,c_cn_t,c_cn_n,c_crdot_r,c_crdot_t,c_crdot_n,c_crdot_rdot,c_ctdot_r,c_ctdot_t,c_ctdot_n,c_ctdot_rdot,c_ctdot_tdot,c_cndot_r,c_cndot_t,c_cndot_n,c_cndot_rdot,c_cndot_tdot,c_cndot_ndot,c_cdrg_r,c_cdrg_t,c_cdrg_n,c_cdrg_rdot,c_cdrg_tdot,c_cdrg_ndot,c_cdrg_drg,c_csrp_r,c_csrp_t,c_csrp_n,c_csrp_rdot,c_csrp_tdot,c_csrp_ndot,c_csrp_drg,c_csrp_srp,c_cthr_r,c_cthr_t,c_cthr_n,c_cthr_rdot,c_cthr_tdot,c_cthr_ndot,c_cthr_drg,c_cthr_srp,c_cthr_thr,jspoc_cdm_id,t_span,c_span
0,12345,2018-08-26 09:21:01.000000,,1,2016-05-16 08:08:45.000000,ABC12345,ABC12345,ABC12345,,2016-05-16 08:08:45.000000,0.834492808294835,0.124632824264698,0.80069109437128,0.561214665433106,0.816825630572017,0.208607555206536,0.731592025459866,0.570303067220067,,,,,,,,,,0.1,0.458882862383152,0,,,1234,ABC12345,ABC12345,ABC12345,NONE,CALCULATED,YES,ITRF,PAYLOAD,2,,ABC12345,ABC12345,"MOON,SUN",True,True,False,,,2016-05-16 08:08:45.000000,2016-05-16 08:08:45.000000,0.458882862383152,0.458882862383152,0.458882862383152,0.458882862383152,,,0.458882862383152,0.458882862383152,ABC12345,0.458882862383152,,,,0.458882862383152,0.458882862383152,0,0.458882862383152,,0.25498437500894,0.112376563676326,0.118703524110389,0.452650747012971,0.214489162297546,0.803282382755525,0.786085040093702,0.990171633378862,0.307736824573957,0.50965140364303,0.714555349450332,0.984888473667389,,0.711134792929336,0.198921153455887,0.831195371950245,0.954944212000877,0.704534234094459,0.249467464608224,0.853648472746269,0.415487189920632,0.908484946608889,0.205790492133707,0.955441709404618,0.443781744705242,0.934633773841681,0.113669321432054,0.346558342029125,0.999943982577824,0.260165052897464,0.534910595773655,0.003138518289742,0.645058510230789,0.665414786995049,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,,,,,,,,,,,1234,ABC12345,ABC12345,ABC12345,NONE,CALCULATED,,ITRF,DEBRIS,1,,ABC12345,ABC12345,"MOON,SUN",True,True,False,,,2016-05-16 08:08:45.000000,2016-05-16 08:08:45.000000,0.324514175884396,0.324514175884396,0.324514175884396,0.324514175884396,,,0.324514175884396,0.324514175884396,ABC12345,0.324514175884396,,,,0.324514175884396,0.324514175884396,0,0.324514175884396,,0.758968377374672,0.599509570566248,0.541702597404864,0.657962800075275,0.133389637846314,0.314389472287636,0.172063653702484,0.680568231092825,0.118929030153042,0.402688316488102,0.241770108925465,0.955386508712671,,0.787782142485196,0.683142973728186,0.723077873320779,0.73410866374641,0.699027719390639,0.034462566703314,0.559550929813111,0.905949292244197,0.125685830718603,0.70393549649449,0.331441271306996,0.600239238337988,0.651717098345195,0.169477542016838,0.635691408775896,0.289353669273576,0.35808862247228,0.030280326018385,0.908101381313104,0.068303959763316,0.214876608923229,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,,,,,,,,,,12345,11,1
1,12345,2018-08-26 06:51:01.000000,MAN,1,2016-05-16 08:08:45.000000,ABC12345,ABC12345,,,2016-05-16 08:08:45.000000,0.89873750846093,0.628380794238599,0.277830602478051,0.388502510320027,0.947139473062045,0.288681055107551,0.330496807664414,0.579047626342338,,,RTN,ELLIPSOID,1000,1000,1000,,,0.1,0.458882862383152,,,,1234,ABC12345,ABC12345,ABC12345,NONE,CALCULATED,,EME2000,PAYLOAD,,,,,,,,,,,,,,,,,,,,,,0.458882862383152,,,0.458882862383152,,,,,,0.458882862383152,0.619593868338859,0.280361151863788,0.517707153564549,0.714241233692547,0.896372753026402,0.608950551433962,0.610078567260023,0.99595085102525,0.437724979814095,0.783441124769422,0.329659907491579,,0.999480876708531,0.0427529023373,0.708967665008883,0.384518308333597,0.304580463020946,0.671728496068814,0.032383339709698,0.465593444565189,0.652971618884818,0.060653418516399,0.59523840375538,0.724882649734235,0.267616325965723,0.234322500028149,0.476667920151508,0.398651489942665,0.57705976320159,0.378791681290921,0.704649407619613,0.388876464311396,0.66124248082016,,,,,,,,,,,,,,,,,,,,,,,,,,1234,ABC12345,ABC12345,ABC12345,NONE,CALCULATED,,EME2000,DEBRIS,,,,,,,,,ABC12345,,,,,,,,,,,,,0.324514175884396,,,0.324514175884396,,,,,,0.324514175884396,0.902389838589119,0.814302547930863,0.741191579850528,0.38587542986633,0.642231309351491,0.906550908711897,0.123148444539083,0.650368061638809,0.010587221576897,0.854000303273469,0.684061564016819,,0.632567404007586,0.075007806550711,0.592565527585434,0.749688928299355,0.50972428079729,0.574285445943895,0.969742310001293,0.737128177440502,0.107949278907619,0.290943703797091,0.703043092143929,0.949642846878678,0.064790432064323,0.72636822759012,0.230985476177261,0.136556355763866,0.460149398175066,0.982745545187407,0.918595138753813,0.346770256028981,0.790935062204615,,,,,,,,,,,,,,,,,,,,,,,,,,12,1
2,12346,2018-08-26 06:51:01.000000,MAN,1,2016-05-16 08:08:45.000000,ABC12345,ABC12345,,,2016-05-16 08:08:45.000000,0.89873750846093,0.628380794238599,0.277830602478051,0.388502510320027,0.947139473062045,0.288681055107551,0.330496807664414,0.579047626342338,,,RTN,ELLIPSOID,1000,1000,1000,,,0.1,0.458882862383152,,,,1234,ABC12345,ABC12345,ABC12345,NONE,CALCULATED,,EME2000,PAYLOAD,,,,,,,,,,,,,,,,,,,,,,0.458882862383152,,,0.458882862383152,,,,,,0.458882862383152,0.619593868338859,0.280361151863788,0.517707153564549,0.714241233692547,0.896372753026402,0.608950551433962,0.610078567260023,0.99595085102525,0.437724979814095,0.783441124769422,0.329659907491579,,0.999480876708531,0.0427529023373,0.708967665008883,0.384518308333597,0.304580463020946,0.671728496068814,0.032383339709698,0.465593444565189,0.652971618884818,0.060653418516399,0.59523840375538,0.724882649734235,0.267616325965723,0.234322500028149,0.476667920151508,0.398651489942665,0.57705976320159,0.378791681290921,0.704649407619613,0.388876464311396,0.66124248082016,,,,,,,,,,,,,,,,,,,,,,,,,,1234,ABC12345,ABC12345,ABC12345,NONE,CALCULATED,,EME2000,DEBRIS,,,,,,,,,ABC12345,,,,,,,,,,,,,0.324514175884396,,,0.324514175884396,,,,,,0.324514175884396,0.902389838589119,0.814302547930863,0.741191579850528,0.38587542986633,0.642231309351491,0.906550908711897,0.123148444539083,0.650368061638809,0.010587221576897,0.854000303273469,0.684061564016819,,0.632567404007586,0.075007806550711,0.592565527585434,0.749688928299355,0.50972428079729,0.574285445943895,0.969742310001293,0.737128177440502,0.107949278907619,0.290943703797091,0.703043092143929,0.949642846878678,0.064790432064323,0.72636822759012,0.230985476177261,0.136556355763866,0.460149398175066,0.982745545187407,0.918595138753813,0.346770256028981,0.790935062204615,,,,,,,,,,,,,,,,,,,,,,,,,,12,1
3,12347,2018-08-26 06:51:01.000000,MAN,1,2016-05-16 08:08:45.000000,ABC12345,ABC12345,,,2016-05-16 08:08:45.000000,0.89873750846093,0.628380794238599,0.277830602478051,0.388502510320027,0.947139473062045,0.288681055107551,0.330496807664414,0.579047626342338,,,RTN,ELLIPSOID,1000,1000,1000,,,0.1,0.458882862383152,,,,1234,ABC12345,ABC12345,ABC12345,NONE,CALCULATED,,EME2000,PAYLOAD,,,,,,,,,,,,,,,,,,,,,,0.458882862383152,,,0.458882862383152,,,,,,0.458882862383152,0.619593868338859,0.280361151863788,0.517707153564549,0.714241233692547,0.896372753026402,0.608950551433962,0.610078567260023,0.99595085102525,0.437724979814095,0.783441124769422,0.329659907491579,,0.999480876708531,0.0427529023373,0.708967665008883,0.384518308333597,0.304580463020946,0.671728496068814,0.032383339709698,0.465593444565189,0.652971618884818,0.060653418516399,0.59523840375538,0.724882649734235,0.267616325965723,0.234322500028149,0.476667920151508,0.398651489942665,0.57705976320159,0.378791681290921,0.704649407619613,0.388876464311396,0.66124248082016,,,,,,,,,,,,,,,,,,,,,,,,,,1234,ABC12345,ABC12345,ABC12345,NONE,CALCULATED,,EME2000,DEBRIS,,,,,,,,,ABC12345,,,,,,,,,,,,,0.324514175884396,,,0.324514175884396,,,,,,0.324514175884396,0.902389838589119,0.814302547930863,0.741191579850528,0.38587542986633,0.642231309351491,0.906550908711897,0.123148444539083,0.650368061638809,0.010587221576897,0.854000303273469,0.684061564016819,,0.632567404007586,0.075007806550711,0.592565527585434,0.749688928299355,0.50972428079729,0.574285445943895,0.969742310001293,0.737128177440502,0.107949278907619,0.290943703797091,0.703043092143929,0.949642846878678,0.064790432064323,0.72636822759012,0.230985476177261,0.136556355763866,0.460149398175066,0.982745545187407,0.918595138753813,0.346770256028981,0.790935062204615,,,,,,,,,,,,,,,,,,,,,,,,,,12,1
4,12346,2018-08-26 06:51:01.000000,MAN,1,2016-05-16 08:08:45.000000,ABC12345,ABC12345,,,2016-05-16 08:08:45.000000,0.89873750846093,0.628380794238599,0.277830602478051,0.388502510320027,0.947139473062045,0.288681055107551,0.330496807664414,0.579047626342338,,,RTN,ELLIPSOID,1000,1000,1000,,,0.1,0.458882862383152,,,,1234,ABC12345,ABC12345,ABC12345,NONE,CALCULATED,,EME2000,PAYLOAD,,,,,,,,,,,,,,,,,,,,,,0.458882862383152,,,0.458882862383152,,,,,,0.458882862383152,0.619593868338859,0.280361151863788,0.517707153564549,0.714241233692547,0.896372753026402,0.608950551433962,0.610078567260023,0.99595085102525,0.437724979814095,0.783441124769422,0.329659907491579,,0.999480876708531,0.0427529023373,0.708967665008883,0.384518308333597,0.304580463020946,0.671728496068814,0.032383339709698,0.465593444565189,0.652971618884818,0.060653418516399,0.59523840375538,0.724882649734235,0.267616325965723,0.234322500028149,0.476667920151508,0.398651489942665,0.57705976320159,0.378791681290921,0.704649407619613,0.388876464311396,0.66124248082016,,,,,,,,,,,,,,,,,,,,,,,,,,1234,ABC12345,ABC12345,ABC12345,NONE,CALCULATED,,EME2000,DEBRIS,,,,,,,,,ABC12345,,,,,,,,,,,,,0.324514175884396,,,0.324514175884396,,,,,,0.324514175884396,0.902389838589119,0.814302547930863,0.741191579850528,0.38587542986633,0.642231309351491,0.906550908711897,0.123148444539083,0.650368061638809,0.010587221576897,0.854000303273469,0.684061564016819,,0.632567404007586,0.075007806550711,0.592565527585434,0.749688928299355,0.50972428079729,0.574285445943895,0.969742310001293,0.737128177440502,0.107949278907619,0.290943703797091,0.703043092143929,0.949642846878678,0.064790432064323,0.72636822759012,0.230985476177261,0.136556355763866,0.460149398175066,0.982745545187407,0.918595138753813,0.346770256028981,0.790935062204615,,,,,,,,,,,,,,,,,,,,,,,,,,12,1

"""
        file_name = os.path.join(tempfile.mkdtemp(), str(uuid.uuid4()))
        with open(file_name, 'w') as f:
            f.write(file_content)

        df = pd.read_csv(file_name)  # Read the csv

        # Make EventDataset from Pandas DataFrame
        e = EventDataset.from_pandas(df)

        # Make Pandas DataFrame from EventDataset
        df = e.to_dataframe()
        self.assertTrue(True)
